import logging
from typing import Dict, List, Optional

import torch
from allennlp.data import Vocabulary
from allennlp.models import Model
from allennlp.modules.transformer import TransformerEmbeddings, TransformerStack, TransformerPooler
from torch.nn import Dropout

logger = logging.getLogger(__name__)


@Model.register("transformer_mc_tt")
class TransformerMCTransformerToolkit(Model):
    """
    This class implements a multiple choice model patterned after the proposed model in
    [RoBERTa: A Robustly Optimized BERT Pretraining Approach (Liu et al)]
    (https://api.semanticscholar.org/CorpusID:198953378).

    It is exactly like the `TransformerMC` model, except it uses the `TransformerTextField` for its input.

    It calculates a score for each sequence on top of the CLS token, and then chooses the alternative
    with the highest score.

    Parameters
    ----------
    vocab : ``Vocabulary``
    transformer_model : ``str``, optional (default=``"roberta-large"``)
        This model chooses the embedder according to this setting. You probably want to make sure this matches the
        setting in the reader.
    """

    def __init__(
        self,
        vocab: Vocabulary,
        transformer_model: str = "roberta-large",
        override_weights_file: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(vocab, **kwargs)
        transformer_kwargs = {
            "model_name": transformer_model,
            "weights_path": override_weights_file,
        }
        self.embeddings = TransformerEmbeddings.from_pretrained_module(**transformer_kwargs)
        self.transformer_stack = TransformerStack.from_pretrained_module(**transformer_kwargs)
        self.pooler = TransformerPooler.from_pretrained_module(**transformer_kwargs)
        self.pooler_dropout = Dropout(p=0.1)

        self.linear_layer = torch.nn.Linear(self.pooler.get_output_dim(), 1)
        self.linear_layer.weight.data.normal_(mean=0.0, std=0.02)
        self.linear_layer.bias.data.zero_()

        self.loss = torch.nn.CrossEntropyLoss()

        from allennlp.training.metrics import CategoricalAccuracy

        self.accuracy = CategoricalAccuracy()

    def forward(  # type: ignore
        self,
        alternatives: Dict[str, torch.Tensor],
        correct_alternative: Optional[torch.IntTensor] = None,
        qid: Optional[List[str]] = None,
    ) -> Dict[str, torch.Tensor]:

        """
        Parameters
        ----------
        alternatives : ``Dict[str, torch.LongTensor]``
            From a ``ListField[TensorTextField]``. Contains a list of alternatives to evaluate for every instance.
        correct_alternative : ``Optional[torch.IntTensor]``
            From an ``IndexField``. Contains the index of the correct answer for every instance.
        qid : `Optional[List[str]]`
            A list of question IDs for the questions being processed now.

        Returns
        -------
        An output dictionary consisting of:
        loss : ``torch.FloatTensor``, optional
            A scalar loss to be optimised. This is only returned when `correct_alternative` is not `None`.
        logits : ``torch.FloatTensor``
            The logits for every possible answer choice
        best_alternative : ``List[int]``
            The index of the highest scoring alternative for every instance in the batch
        """
        batch_size, num_alternatives, seq_length = alternatives["input_ids"].size()

        alternatives = {
            name: t.view(batch_size * num_alternatives, seq_length)
            for name, t in alternatives.items()
        }

        embedded_alternatives = self.embeddings(**alternatives)
        embedded_alternatives = self.transformer_stack(
            embedded_alternatives, alternatives["attention_mask"]
        )
        embedded_alternatives = self.pooler(embedded_alternatives.final_hidden_states)
        embedded_alternatives = self.pooler_dropout(embedded_alternatives)
        logits = self.linear_layer(embedded_alternatives)
        logits = logits.view(batch_size, num_alternatives)

        result = {"logits": logits, "best_alternative": logits.argmax(1)}

        if correct_alternative is not None:
            correct_alternative = correct_alternative.squeeze(1)
            result["loss"] = self.loss(logits, correct_alternative)
            self.accuracy(logits, correct_alternative)

        return result

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        return {
            "acc": self.accuracy.get_metric(reset),
        }



import itertools
import logging
from typing import List, Optional

import torch
from allennlp.common import cached_transformers
from allennlp.data import DatasetReader, Instance
from allennlp.data.fields import TransformerTextField
from overrides import overrides

logger = logging.getLogger(__name__)


class TransformerMCReaderTransformerToolkit(DatasetReader):
    """
    Read input data for the TransformerMC model. This is the base class for all readers that produce
    data for TransformerMCTransformerToolkit.

    Instances have three fields:
     * `alternatives`, a `ListField` of `TransformerTextField`
     * `correct_alternative`, `IndexField` with the correct answer among `alternatives`
     * `qid`, a `MetadataField` containing question ids

    Parameters
    ----------
    transformer_model_name : `str`, optional (default=`"roberta-large"`)
        This reader chooses tokenizer and token indexer according to this setting.
    length_limit : `int`, optional (default=`512`)
        We will make sure that the length of an alternative never exceeds this many word pieces.
    """

    def __init__(
        self, transformer_model_name: str = "roberta-large", length_limit: int = 512, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self._tokenizer = cached_transformers.get_tokenizer(transformer_model_name)
        self.length_limit = length_limit

    @overrides
    def text_to_instance(
        self,  # type: ignore
        qid: str,
        start: str,
        alternatives: List[str],
        label: Optional[int] = None,
    ) -> Instance:
        start = start.strip()

        tokenized = self._tokenizer(
            [(start, a) for a in alternatives],
            truncation="longest_first",
            max_length=self.length_limit,
            return_attention_mask=False,
        )
        sequences = [
            TransformerTextField(
                torch.IntTensor(input_ids),
                torch.IntTensor(token_type_ids) if token_type_ids is not None else None,
                padding_token_id=self._tokenizer.pad_token_id,
            )
            for input_ids, token_type_ids in itertools.zip_longest(
                tokenized["input_ids"], tokenized.get("token_type_ids", [])
            )
        ]

        from allennlp.data.fields import ListField

        sequences = ListField(sequences)

        from allennlp.data.fields import MetadataField

        fields = {
            "alternatives": sequences,
            "qid": MetadataField(qid),
        }

        if label is not None:
            if label < 0 or label >= len(sequences):
                raise ValueError("Alternative %d does not exist", label)
            from allennlp.data.fields import IndexField

            fields["correct_alternative"] = IndexField(label, sequences)

        return Instance(fields)



import logging

from allennlp.data import DatasetReader

from allennlp_models.mc.dataset_readers.piqa import PiqaReader
from allennlp_models.mc.dataset_readers.transformer_mc_tt import (
    TransformerMCReaderTransformerToolkit,
)

logger = logging.getLogger(__name__)


@DatasetReader.register("piqa_tt")
class PiqaReaderTransformerToolkit(TransformerMCReaderTransformerToolkit, PiqaReader):
    pass
